
from torch.utils.data import Dataset
import torch
import numpy as np
import os
import re
from skimage import exposure
from copy import deepcopy
from tqdm import tqdm

class cellsDataset(Dataset):
    def __init__(self, images_path, somas_path, balance_data = False, transform=None, label=True, hist_equal=False, load2Ram=False):
        self.images_path = images_path
        self.somas_path = somas_path
        self.transform = transform
        self.balance_data = balance_data
        self.label = label
        self.hist_equal = hist_equal
        self.load2Ram = load2Ram
        self.make_paths()
        if self.load2Ram:
            self.load_all()
        

    def make_paths(self):
        self.images = os.listdir(self.images_path)
        if self.label:
            self.somas = os.listdir(self.somas_path)
        self.images.sort(key=lambda f: int(re.sub('\D', '', f)))
        if self.label:
            self.somas.sort(key=lambda f: int(re.sub('\D', '', f)))
        self.images = [os.path.join(self.images_path, img) for img in self.images]
        if self.label:
            self.somas = [os.path.join(self.somas_path, soma) for soma in self.somas]
            

    def __len__(self):
        return len(self.images)
    

    def load_all(self):
        self.samples = []
        print('Loading the data into the RAM')
        for idx in tqdm(range(len(self.images))):
            image_paths = self.images[idx]
            if self.label:
                soma_paths = self.somas[idx]

            image = np.load(image_paths)
            if self.label:
                soma = np.load(soma_paths)

            if self.hist_equal:
                img = deepcopy(image['arr_0'])
                img = exposure.equalize_adapthist(np.squeeze(np.array(image['arr_0'])),5)*255
            else:
                img = image['arr_0']

            if self.label:
                sample = {'image': img, 'soma': soma['arr_0']}
            else:
                sample = {'image': image['arr_0']}
            
            sample['image'] = np.expand_dims(sample['image'], 0) # Batch Size = 1
            if self.label:
                sample['soma'] = np.expand_dims(sample['soma'], 0)

            sample['image'] = torch.tensor(sample['image'], dtype=torch.float32)
            if self.label:
                sample['soma'] = torch.tensor(sample['soma'], dtype=torch.bool)
            

            self.samples.append(sample) 


    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
            
        if self.load2Ram:
            sample = self.samples[idx]
        else:
            image_paths = self.images[idx]
            if self.label:
                soma_paths = self.somas[idx]

            image = np.load(image_paths)
            if self.label:
                soma = np.load(soma_paths)

            if self.hist_equal:
                img = deepcopy(image['arr_0'])
                img = exposure.equalize_adapthist(np.squeeze(np.array(image['arr_0'])),5)*255
            else:
                img = image['arr_0']

            if self.label:
                sample = {'image': img, 'soma': soma['arr_0']}
            else:
                sample = {'image': image['arr_0']}
            
            sample['image'] = np.expand_dims(sample['image'], 0) # Batch Size = 1
            if self.label:
                sample['soma'] = np.expand_dims(sample['soma'], 0)

            sample['image'] = torch.tensor(sample['image'], dtype=torch.float32)
            if self.label:
                sample['soma'] = torch.tensor(sample['soma'], dtype=torch.bool)


        if self.transform and self.label:
            sample = self.transform(sample)

        return sample


import numpy as np
from tqdm import tqdm
import torch

def reconstruct_image(imgs, metadata):
    img_size = metadata['img_size']
    window_size_x = metadata['window_size_x']
    window_size_y = metadata['window_size_y']
    window_size_z = metadata['window_size_z']
    overlap_x = metadata['overlap_x']
    overlap_y = metadata['overlap_y']
    overlap_z = metadata['overlap_z']
    img_all = np.zeros(img_size, dtype=np.uint8)
    print(img_size, int(img_size[0]/(window_size_x-overlap_x)))
    counter = 0
    for i_counter in tqdm(range(0, int(img_size[0]/(window_size_x-overlap_x))+1)):
        for j_counter in range(0, int(img_size[1]/(window_size_y-overlap_y))+1):
            for k_counter in range(0, int(img_size[2]/(window_size_z-overlap_z))+1):
                start_x = i_counter*(window_size_x-overlap_x)
                start_y = j_counter*(window_size_y-overlap_y)
                start_z = k_counter*(window_size_z-overlap_z)
                end_x = start_x+window_size_x
                end_y = start_y+window_size_y
                end_z = start_z+window_size_z

                if end_x > img_size[0]:
                    end_x = img_size[0]
                    start_x = end_x - window_size_x
                if end_y > img_size[1]:
                    end_y = img_size[1]
                    start_y = end_y - window_size_y
                
                if end_z > img_size[2]:
                    end_z = img_size[2]
                    start_z = end_z - window_size_z
                img_all[start_x:end_x, start_y:end_y, start_z:end_z] = imgs[counter, : , :, :]
                counter += 1
    return img_all

def reconstruct_images_no_overlap(imgs):
    num_imgs = imgs.shape[0]
    num_rows = int(np.sqrt(num_imgs))
    num_cols = int(np.ceil(num_imgs/num_rows))

    img_size = imgs.shape[1:]
    img_all = np.zeros((num_rows*img_size[0], num_cols*img_size[1], img_size[2]), dtype=np.float32)

    for row_counter in range(num_rows):
        for col_counter in range(num_cols):
            img_counter = row_counter*num_cols + col_counter
            if img_counter >= num_imgs:
                break

            start_x = row_counter*img_size[0]
            start_y = col_counter*img_size[1]

            end_x = start_x + img_size[0]
            end_y = start_y + img_size[1]
            img_all[start_x:end_x, start_y:end_y, :] = imgs[img_counter, :, :, :]
    
    return img_all


def convert_soma_and_traces_to_image(cells, img, convert_x, convert_y, convert_z, SizeX, SizeY, SizeZ):
    mask_soma = np.zeros(img.shape)
    mask_traces = np.zeros(img.shape)
    for cell in tqdm(cells.values()):
        traces = cell['traces']
        soma_pos = cell['soma_pos']
        x_soma = convert_x(soma_pos[0])
        y_soma = convert_y(soma_pos[1])
        z_soma = convert_z(soma_pos[2])
        mask_soma[x_soma, y_soma, z_soma] = 1
        kernel_size = 3
        # convolve soma with the kernel
        kernel = np.ones((kernel_size, kernel_size, kernel_size))
        mask_soma = scipy.signal.convolve(mask_soma, kernel, mode='same')

        for trace_row in traces:
            x = convert_x(trace_row[0])
            y = convert_y(trace_row[1])
            z = convert_z(trace_row[2])
            if x >= SizeX:
                Warning(f'x out of range. Should be less than {SizeX}, but got {x}')
                x = SizeX-1
            if y >= SizeY:
                Warning(f'y out of range. Should be less than {SizeY}, but got {y}')
                y = SizeY-1
            if z >= SizeZ:
                Warning(f'z out of range. Should be less than {SizeZ}, but got {z}')
                z = SizeZ-1
            mask_traces[x, y, z] = 1
    
    return mask_soma, mask_traces


def prepare_image_for_plot(image):
    image = (image - np.min(image))/(np.max(image) - np.min(image))
    image = np.uint8(image*255)
    return image

def get_weights_of_the_model(model):
    weights = []
    for name, param in model.named_parameters():
        if 'weight' in name:
            weights.append(param)
    weights = torch.cat([w.view(-1) for w in weights])
    return weights

def normalize_img(x):
    x = x-torch.min(x)
    x = x/torch.max(x)
    return x


import torch
import torch.nn as nn

class CellSomaSegmentationModel(nn.Module):
    def __init__(self):
        super(CellSomaSegmentationModel, self).__init__()
        self.dropout_prob = 0.3
        self.num_features = 16
        
        # Encoder
        self.down_1 = nn.Sequential(
            # input size 1*128x128x16
            nn.Conv3d(1, self.num_features, kernel_size=(8,8,2), padding=(3,3,0), stride=(2,2,2)),
            nn.BatchNorm3d(self.num_features),
            nn.ReLU(),
            nn.Dropout3d(self.dropout_prob),
            # output size 8*64x64x8
        )
        self.down_2 = nn.Sequential(
            # input size 8*64x64x8
            nn.Conv3d(self.num_features, 2*self.num_features, kernel_size=(8,8,3), padding=(3,3,1), stride=(2,2,1)),
            nn.BatchNorm3d(2*self.num_features),
            nn.ReLU(),
            nn.Dropout3d(self.dropout_prob),
            # output size 16*32x32x8
        )
        self.down_3 = nn.Sequential(
            # input size 16*32x32x8
            nn.Conv3d(2*self.num_features, 4*self.num_features, kernel_size=(8,8,3), padding=(3,3,1), stride=(2,2,1)),
            nn.BatchNorm3d(4*self.num_features),
            nn.ReLU(),
            # output size 32*16x16x8
        )

        self.double_conv_1 = nn.Sequential(
            # input size 32*16x16x8
            nn.Conv3d(4*self.num_features, 4*self.num_features, kernel_size=(3,3,3), padding=(1,1,1), stride=(1,1,1)),
            nn.ReLU(),
            nn.Dropout3d(self.dropout_prob),
            nn.Conv3d(4*self.num_features, 4*self.num_features, kernel_size=(3,3,3), padding=(1,1,1), stride=(1,1,1)),
            nn.ReLU(),
            nn.Dropout3d(self.dropout_prob),
            # output size 32*16x16x8
        )

        # Decoder
        self.up_1 = nn.Sequential(
            # input size 64*16x16x8
            nn.ConvTranspose3d(8*self.num_features, 4*self.num_features, kernel_size=(8,8,3), padding=(3,3,1), stride=(2,2,1)),
            nn.BatchNorm3d(4*self.num_features),
            nn.ReLU(),
            nn.Dropout3d(self.dropout_prob),
            # output size 32*32x32x8
        )
        self.up_2 = nn.Sequential(
            # input size 48*32x32x8
            nn.ConvTranspose3d(6*self.num_features, 2*self.num_features, kernel_size=(8,8,3), padding=(3,3,1), stride=(2,2,1)),
            nn.BatchNorm3d(2*self.num_features),
            nn.ReLU(),
            nn.Dropout3d(self.dropout_prob),
            # output size 16*64x64x8
        )
        self.up_3 = nn.Sequential(
            # input size 24*64x64x8
            nn.ConvTranspose3d(3*self.num_features, 1, kernel_size=(8,8,4), padding=(3,3,1), stride=(2,2,2)),
            nn.BatchNorm3d(1),
            nn.Hardtanh(min_val=0, max_val=1),
            # output size 1*128x128x16
        )
    
    def forward(self, x):
        # x shape is 1*64x64x16
        x1 = self.down_1(x)
        # x1 shape is 8*32x32x8
        x2 = self.down_2(x1)
        # x2 shape is 16*16x16x8
        x3 = self.down_3(x2)
        # x3 shape is 32*8x8x8
        x4 = self.double_conv_1(x3)
        # x4 shape is 32*8x8x8
        x5 = torch.cat([x4, x3], dim=1)
        # x5 shape is 64*8x8x8
        x6 = self.up_1(x5)
        # x6 shape is 32*16x16x8
        x7 = torch.cat([x6, x2], dim=1)
        # x7 shape is 48*16x16x8
        x8 = self.up_2(x7)
        # x8 shape is 16*32x32x8
        x9 = torch.cat([x8, x1], dim=1)
        # x9 shape is 24*32x32x16
        x10 = self.up_3(x9)
        # x10 shape is 1*64x64x16

        return x10


# import cv2
import torch
import torch.nn as nn
from torch.utils.data import DataLoader, Dataset
import torch.nn.functional as F
import torch.optim as optim
import torchvision.transforms as transforms
import matplotlib.pyplot as plt
import os
import mat73
import numpy as np
import json
import torchio as tio
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
from tqdm import tqdm
import torch.nn.functional as F
from copy import deepcopy, copy
import tifffile as tiff
from scipy import ndimage
import click
from math import floor
from scipy.ndimage import grey_dilation

from models import CellSomaSegmentationModel

from utils import *
from datasets import cellsDataset

@click.group(chain=True)
def cli():
    pass

@cli.command('RGC_detection')
@click.option('--file_path', '-fp', type=str, required=True, help='Path to the image file or directory containing image files (.ims files supported for now)')
def fluer_pipeline(file_path):
    if os.path.isfile(file_path):
        file_paths = [file_path]
    else:
        file_paths = [os.path.join(file_path, f) for f in os.listdir(file_path) if os.path.isfile(os.path.join(file_path, f)) and f.endswith('.ims')]
    
    for file_path in file_paths:
        print(f"file_path: {file_path}")

        img_path = file_path
        dataset_name = os.path.basename(img_path).split(".")[0]
        print(f"dataset_name: {dataset_name}")
        mat_path = os.path.join("tmp", dataset_name, 'mat')
        save_path_dataset = os.path.join("tmp", dataset_name, 'dataset')

        print(f"img_path: {img_path}")
        print("=======Make mat fules out of img/nd2 file=======")

        if not os.path.exists(save_path_dataset):
            os.makedirs(save_path_dataset)
        
        print(f"=======Make Masks Soma=======")
        datasets_path_train = save_path_dataset 
        num_epochs = 1000
        print(num_epochs)

        load2Ram = False
        hist_equal = False
        transform = None

        datasets_paths = [os.path.join(datasets_path_train, folder) for folder in os.listdir(datasets_path_train) if os.path.isdir(os.path.join(datasets_path_train, folder))]
        print(datasets_paths)
        datasets_list_train = []
        print("128")
        for dataset_path in tqdm(datasets_paths):
            img_datapath = os.path.join(dataset_path, 'img')
            soma_datapath = os.path.join(dataset_path, 'soma')
            metadata_path = os.path.join(dataset_path, 'metadata.json')
            # try:
            # load metadata (Not used)
            with open(metadata_path) as f:
                metadata = json.load(f)
            print(metadata)
            print("124")

            cells_dataset = cellsDataset(img_datapath, soma_datapath, transform=transform, hist_equal=hist_equal, load2Ram=load2Ram, label=False)
            
            datasets_list_train.append(cells_dataset)
            print(f'Loading dataset: {dataset_path} ==> Loaded')
            # except:
            #     print(f'Loading dataset: {dataset_path} ==> Failed')


        # concatenate all datasets
        dataset_train = torch.utils.data.ConcatDataset(datasets_list_train)

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        print(f"Device: {device}")

        model = CellSomaSegmentationModel()
        model = nn.DataParallel(model)
        model.load_state_dict(torch.load(os.path.join('./models/soma_detection.pth')))
        model.eval()
        model.to(device)

        len_dataset = floor(len(dataset_train))
        size_sample = np.squeeze(dataset_train.__getitem__(0)['image']).shape

        images = np.zeros((len_dataset, size_sample[0], size_sample[1], size_sample[2]), dtype=np.uint8)
        # soma = np.zeros((len_dataset, size_sample[0], size_sample[1], size_sample[2]), dtype=np.uint8)
        segmented_images = np.zeros((len_dataset, size_sample[0], size_sample[1], size_sample[2]), dtype=np.uint8)

        for i in tqdm(range(len_dataset)):
            sample = dataset_train.__getitem__(i)
            images[i, :, :, :] = np.array(sample['image'], dtype=np.uint8)
            # soma[i, :, :, :] = np.array(sample['soma'], dtype=np.uint8)

            segmented = model(torch.tensor(np.expand_dims(sample['image'], 0)).type(torch.float32).to(device))
            segmented = segmented.squeeze(0).cpu().detach().numpy()
            segmented = np.uint8(segmented*255)
            segmented_images[i, :, :, :] = segmented

        images =  reconstruct_image(images, metadata)
        segmented_images =  reconstruct_image(segmented_images, metadata)

        # Create a 3D structuring element
        structuring_element = np.ones((5, 5, 5))

        # Ensure your image is uint8
        segmented_images = segmented_images > 125
        segmented_images = segmented_images.astype(np.uint8)

        # Dilate the imageow
        dilated_image = grey_dilation(segmented_images, footprint=structuring_element)

        _, num_labels = ndimage.label(dilated_image)

        print(f"Number of cells: {num_labels+1}")
        model_save_path_soma = os.path.join("tmp", dataset_name, "soma_out")
        if not os.path.exists(model_save_path_soma):
            os.makedirs(model_save_path_soma, exist_ok=True)
        # Write the number of cells to a text file
        with open(os.path.join(model_save_path_soma,'number_of_cells.txt'), 'w') as file:
            file.write(str(num_labels+1))
        print("Save Starting")

        images = np.transpose(images, [2,0,1])
        segmented_images = np.transpose(segmented_images, [2,0,1])
        for img, name in zip([images,segmented_images], ['img','segmented']): 
            name = f'{name}.tif'
            path = os.path.join(model_save_path_soma, name)
            print(path)
            tiff.imwrite(path, img)

if __name__ == '__main__':
    cli()


import os
from setuptools import setup, find_packages
from distutils.core import Extension
from pathlib import Path

MINIMAL_DESCRIPTION = '''RGC-Quant: Retinal Ganglion Cell (RGC) automatic counting in 3D confocal images using Deep Learning'''

def get_requires():
    """Read requirements.txt."""
    requirements_file = os.path.join(os.path.dirname(__file__), "requirements.txt")
    try:
        with open(requirements_file, "r") as f:
            requirements = f.read()
        return list(filter(lambda x: x != "", requirements.split()))
    except FileNotFoundError:
        return []

def read_description():
    """Read README.md and CHANGELOG.md."""
    readme_path = Path("README.md")
    if readme_path.exists():
        with open(readme_path) as r:
            description = "\n" + r.read()
        return description
    return MINIMAL_DESCRIPTION

setup(
    name="RGC-Quant",
    version="0.0.1",
    author='Amin Alam, Arsalan Firoozi',
    description='RGC Quantification',
    long_description=read_description(),
    long_description_content_type='text/markdown',
    install_requires=get_requires(),
    python_requires='>=3.8',
    license='MIT',
    url='https://github.com/AminAlam/RGC-Quant',
    keywords=['RGC', 'Quantification', 'Deep Learning', '3D', 'Confocal', 'Microscopy', 'Retina', 'Neuroscience', 'Image Processing', 'Image Analysis', 'Image Segmentation', 'Deep learning', 'PyTorch', '3D U-Net'],
    entry_points={
        'console_scripts': [
            'rgc_quant=rgc_quant.main:main',
        ],
    },
    packages=['rgc_quant'],
    package_data={'rgc_quant': ['*.pth', 'rgc_quant/models/*.m']},
    include_package_data=True,
)